from typing import List

import numpy

from pyquda_comm import getLogger
from pyquda_comm.field import LatticeInfo, LatticeGauge, LatticeStaggeredFermion, MultiLatticeStaggeredFermion
from ..pyquda import computeHISQForceQuda, dslashQuda, saveGaugeQuda
from ..enum_quda import (
    QudaInverterType,
    QudaMassNormalization,
    QudaMatPCType,
    QudaParity,
    QudaSolutionType,
    QudaSolveType,
    QudaVerbosity,
)
from ..dirac import HISQDirac

nullptr = numpy.empty((0, 0), "<c16")

from .abstract import RationalParam, StaggeredFermionAction


class HISQAction(StaggeredFermionAction):
    dirac: HISQDirac

    def __init__(
        self,
        latt_info: LatticeInfo,
        rational_param: RationalParam,
        tol: float,
        maxiter: int,
        naik_epsilon: float = 0.0,
        verbosity: QudaVerbosity = QudaVerbosity.QUDA_SILENT,
    ) -> None:
        if latt_info.anisotropy != 1.0:
            getLogger().critical("anisotropy != 1.0 not implemented", NotImplementedError)
        super().__init__(latt_info, HISQDirac(latt_info, 0.0, tol, maxiter, naik_epsilon, None))

        self.setForceParam(rational_param)
        self.quark = None  # For MultiHISQAction
        self.phi = LatticeStaggeredFermion(latt_info)
        self.eta = LatticeStaggeredFermion(latt_info)

        self.invert_param.inv_type = QudaInverterType.QUDA_CG_INVERTER
        self.invert_param.solution_type = QudaSolutionType.QUDA_MATPC_SOLUTION
        self.invert_param.solve_type = QudaSolveType.QUDA_DIRECT_PC_SOLVE
        self.invert_param.matpc_type = QudaMatPCType.QUDA_MATPC_EVEN_EVEN
        self.invert_param.mass_normalization = QudaMassNormalization.QUDA_MASS_NORMALIZATION
        self.invert_param.verbosity = verbosity

    def setForceParam(self, rational_param: RationalParam):
        coeff = [
            [
                2 * res,
                self.dirac.path_coeff_2[1] * 2 * res,
            ]
            for res in rational_param.residue_molecular_dynamics
        ]
        if self.dirac.naik_epsilon != 0.0:
            coeff += [
                [
                    self.dirac.path_coeff_3[0] * self.dirac.naik_epsilon * 2 * res,
                    self.dirac.path_coeff_3[1] * self.dirac.naik_epsilon * 2 * res,
                ]
                for res in rational_param.residue_molecular_dynamics
            ]
        self.num = len(rational_param.offset_molecular_dynamics)
        self.num_naik = 0 if self.dirac.naik_epsilon == 0.0 else self.num
        self.coeff = numpy.array(coeff, "<f8")
        self.max_num_offset = max(
            len(rational_param.offset_molecular_dynamics),
            len(rational_param.offset_fermion_action),
            len(rational_param.offset_pseudo_fermion),
        )
        self.rational_param = rational_param

    def updateFatLong(self, return_v_link: bool):
        u_link = LatticeGauge(self.latt_info)
        saveGaugeQuda(u_link.data_ptrs, self.gauge_param)
        v_link, w_link = self.dirac.computeWLink(u_link, return_v_link)
        fatlink, longlink = self.dirac.computeXLink(w_link)
        fatlink, longlink = self.dirac.computeXLinkEpsilon(fatlink, longlink, w_link)
        self.dirac.loadFatLongGauge(fatlink, longlink)

        return u_link, v_link, w_link

    def sample(self, new_gauge: bool):
        if self.quark is None:
            self.quark = MultiLatticeStaggeredFermion(self.latt_info, self.max_num_offset)
        self.sampleEta()
        self.updateFatLong(False)
        self.invertMultiShift("pseudo_fermion")

    def action(self, new_gauge: bool) -> float:
        self.updateFatLong(False)
        self.invert_param.compute_action = 1
        self.invertMultiShift("molecular_dynamics")
        self.invert_param.compute_action = 0
        return self.invert_param.action[0]

    def actionFA(self, new_gauge: bool) -> float:
        self.updateFatLong(False)
        self.invertMultiShift("fermion_action")
        return self.eta.even.norm2()  # - norm_molecular_dynamics * self.phi.even.norm2()

    def force(self, dt, new_gauge: bool):
        u_link, v_link, w_link = self.updateFatLong(True)
        self.invertMultiShift("molecular_dynamics")
        for i in range(self.num):
            dslashQuda(self.quark[i].odd_ptr, self.quark[i].even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        computeHISQForceQuda(
            nullptr,
            dt,
            self.dirac.path_coeff_2,
            self.dirac.path_coeff_1,
            w_link.data_ptrs,
            v_link.data_ptrs,
            u_link.data_ptrs,
            self.quark.data_ptrs,
            self.num,
            self.num_naik,
            self.coeff,
            self.gauge_param,
        )


class MultiHISQAction(StaggeredFermionAction):
    dirac: HISQDirac

    def __init__(self, latt_info: LatticeInfo, pseudo_fermions: List[HISQAction]) -> None:
        if latt_info.anisotropy != 1.0:
            getLogger().critical("anisotropy != 1.0 not implemented", NotImplementedError)
        super().__init__(latt_info, pseudo_fermions[0].dirac)

        self.setForceParam(pseudo_fermions)
        self.quark = MultiLatticeStaggeredFermion(self.latt_info, self.max_num_offset)

    def setForceParam(self, pseudo_fermions: List[HISQAction]):
        num_total = 0
        num_naik_total = 0
        for pseudo_fermion in pseudo_fermions:
            num_total += pseudo_fermion.num
            num_naik_total += pseudo_fermion.num_naik
        coeff_all = numpy.zeros((num_total + num_naik_total, 2), "<f8")
        num_current = 0
        num_naik_current = num_total
        for pseudo_fermion in pseudo_fermions:
            num = pseudo_fermion.num
            num_naik = pseudo_fermion.num_naik
            coeff_all[num_current : num_current + num] = pseudo_fermion.coeff[:num]
            coeff_all[num_naik_current : num_naik_current + num_naik] = pseudo_fermion.coeff[num:]
            num_current += num
            num_naik_current += num_naik
        self.num = num_total
        self.num_naik = num_naik_total
        self.coeff = coeff_all
        self.max_num_offset = max([self.num] + [pseudo_fermion.max_num_offset for pseudo_fermion in pseudo_fermions])
        self.pseudo_fermions = pseudo_fermions

    def prepareFatLong(self, return_v_link: bool):
        u_link = LatticeGauge(self.latt_info)
        saveGaugeQuda(u_link.data_ptrs, self.gauge_param)
        v_link, w_link = self.dirac.computeWLink(u_link, return_v_link)
        self.fatlink, self.longlink = self.dirac.computeXLink(w_link)
        self.w_link = w_link

        return u_link, v_link, w_link

    def updateFatLong(self, pseudo_fermion: HISQAction):
        fatlink, longlink = pseudo_fermion.dirac.computeXLinkEpsilon(self.fatlink, self.longlink, self.w_link)
        pseudo_fermion.dirac.loadFatLongGauge(fatlink, longlink)

    def sample(self, new_gauge: bool):
        self.prepareFatLong(False)
        for pseudo_fermion in self.pseudo_fermions:
            pseudo_fermion.sampleEta()
            self.updateFatLong(pseudo_fermion)
            pseudo_fermion.quark = self.quark
            pseudo_fermion.invertMultiShift("pseudo_fermion")

    def action(self, new_gauge: bool) -> float:
        action = 0
        self.prepareFatLong(False)
        for pseudo_fermion in self.pseudo_fermions:
            self.updateFatLong(pseudo_fermion)
            pseudo_fermion.invert_param.compute_action = 1
            pseudo_fermion.quark = self.quark
            pseudo_fermion.invertMultiShift("molecular_dynamics")
            pseudo_fermion.invert_param.compute_action = 0
            action += pseudo_fermion.invert_param.action[0]
        return action

    def actionFA(self, new_gauge: bool) -> float:
        action = 0
        self.prepareFatLong(False)
        for pseudo_fermion in self.pseudo_fermions:
            self.updateFatLong(pseudo_fermion)
            pseudo_fermion.quark = self.quark
            pseudo_fermion.invertMultiShift("fermion_action")
            action += pseudo_fermion.eta.even.norm2()
        return action

    def force(self, dt, new_gauge: bool):
        u_link, v_link, w_link = self.prepareFatLong(True)
        num_current = 0
        for pseudo_fermion in self.pseudo_fermions:
            self.updateFatLong(pseudo_fermion)
            pseudo_fermion.quark = self.quark[num_current : num_current + pseudo_fermion.num]
            pseudo_fermion.invertMultiShift("molecular_dynamics")
            for i in range(pseudo_fermion.num):
                dslashQuda(
                    self.quark[num_current + i].odd_ptr,
                    self.quark[num_current + i].even_ptr,
                    pseudo_fermion.invert_param,
                    QudaParity.QUDA_ODD_PARITY,
                )
            num_current += pseudo_fermion.num
        computeHISQForceQuda(
            nullptr,
            dt,
            self.dirac.path_coeff_2,
            self.dirac.path_coeff_1,
            w_link.data_ptrs,
            v_link.data_ptrs,
            u_link.data_ptrs,
            self.quark.data_ptrs,
            self.num,
            self.num_naik,
            self.coeff,
            self.gauge_param,
        )



from typing import List, Union

from pyquda_comm.field import LatticeInfo, LatticeGauge
from ..enum_quda import QudaDslashType, QudaInverterType, QudaReconstructType, QudaPrecision

from . import general
from .abstract import Multigrid, StaggeredFermionDirac


class StaggeredDirac(StaggeredFermionDirac):
    def __init__(
        self,
        latt_info: LatticeInfo,
        mass: float,
        tol: float,
        maxiter: int,
        tadpole_coeff: float = 1.0,
        multigrid: Union[List[List[int]], Multigrid] = None,
    ) -> None:
        kappa = 1 / 2
        super().__init__(latt_info)
        self.newQudaGaugeParam(tadpole_coeff)
        self.newQudaMultigridParam(multigrid, mass, kappa, 0.25, 16, 1e-6, 1000, 0, 8)
        self.newQudaInvertParam(mass, kappa, tol, maxiter)
        # Using half with multigrid doesn't work
        if multigrid is not None:
            self.setPrecision(sloppy=max(self.precision.sloppy, QudaPrecision.QUDA_SINGLE_PRECISION))
        else:
            self.setPrecision()
        self.setReconstruct(
            cuda=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            sloppy=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            precondition=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            eigensolver=QudaReconstructType.QUDA_RECONSTRUCT_NO,
        )

    def newQudaGaugeParam(self, tadpole_coeff: float):
        gauge_param = general.newQudaGaugeParam(self.latt_info, tadpole_coeff, 0.0)
        gauge_param.staggered_phase_applied = 1
        self.gauge_param = gauge_param

    def newQudaMultigridParam(
        self,
        multigrid: Union[List[List[int]], Multigrid],
        mass: float,
        kappa: float,
        coarse_tol: float,
        coarse_maxiter: int,
        setup_tol: float,
        setup_maxiter: int,
        nu_pre: int,
        nu_post: int,
    ):
        if isinstance(multigrid, Multigrid):
            self.multigrid = multigrid
        elif multigrid is not None:
            geo_block_size = multigrid
            mg_param, mg_inv_param = general.newQudaMultigridParam(
                mass,
                kappa,
                geo_block_size,
                coarse_tol,
                coarse_maxiter,
                setup_tol,
                setup_maxiter,
                nu_pre,
                nu_post,
                True,
            )
            mg_inv_param.dslash_type = QudaDslashType.QUDA_ASQTAD_DSLASH
            self.multigrid = Multigrid(mg_param, mg_inv_param)
        else:
            self.multigrid = Multigrid(None, None)

    def newQudaInvertParam(self, mass: float, kappa: float, tol: float, maxiter: int):
        invert_param = general.newQudaInvertParam(mass, kappa, tol, maxiter, 0.0, 1.0, self.multigrid.param)
        invert_param.dslash_type = QudaDslashType.QUDA_STAGGERED_DSLASH
        if self.multigrid.param is None:
            invert_param.inv_type = QudaInverterType.QUDA_CG_INVERTER
        self.invert_param = invert_param

    def loadGauge(self, gauge: LatticeGauge, thin_update_only: bool = False):
        general.loadStaggeredGauge(gauge, self.gauge_param)
        if self.multigrid.instance is None:
            self.newMultigrid()
        else:
            self.updateMultigrid(thin_update_only)

    def destroy(self):
        self.destroyMultigrid()



from abc import ABC, abstractmethod
from functools import partial
from typing import List, Literal, NamedTuple, Union

from pyquda_comm import getCUDABackend
from pyquda_comm.field import (
    LatticeInfo,
    LatticeFermion,
    LatticeStaggeredFermion,
    MultiLatticeFermion,
    MultiLatticeStaggeredFermion,
)
from ..enum_quda import QUDA_MAX_MULTI_SHIFT, QudaDagType, QudaMatPCType
from ..pyquda import QudaGaugeParam, QudaInvertParam, invertQuda, MatQuda, invertMultiShiftQuda
from ..dirac.abstract import Dirac, FermionDirac, StaggeredFermionDirac


class LoopParam(NamedTuple):
    path: List[List[int]]
    coeff: List[float]


class Action(ABC):
    latt_info: LatticeInfo
    dirac: Dirac
    gauge_param: QudaGaugeParam
    loop_param: LoopParam

    def __init__(self, latt_info: LatticeInfo, dirac: Dirac) -> None:
        self.latt_info = latt_info
        self.dirac = dirac
        self.gauge_param = self.dirac.gauge_param
        self.is_fermion = False
        self.is_staggered = False

    @abstractmethod
    def action(self) -> float:
        pass

    @abstractmethod
    def force(self, dt: float):
        pass


class RationalParam(NamedTuple):
    norm_molecular_dynamics: float = 0.0
    residue_molecular_dynamics: List[float] = [1.0]
    offset_molecular_dynamics: List[float] = [0.0]
    norm_pseudo_fermion: float = 0.0
    residue_pseudo_fermion: List[float] = [1.0]
    offset_pseudo_fermion: List[float] = [0.0]
    norm_fermion_action: float = 0.0
    residue_fermion_action: List[float] = [1.0]
    offset_fermion_action: List[float] = [0.0]


class FermionAction(Action):
    dirac: FermionDirac
    invert_param: QudaInvertParam
    rational_param: RationalParam
    quark: MultiLatticeFermion
    phi: LatticeFermion
    eta: LatticeFermion

    def __init__(self, latt_info: LatticeInfo, dirac: FermionDirac) -> None:
        super().__init__(latt_info, dirac)
        self.invert_param = self.dirac.invert_param
        self.is_fermion = True

    def sampleEta(self):
        def _backend():
            backend = getCUDABackend()
            if backend == "numpy":
                import numpy

                return numpy, numpy.random.random
            elif backend == "cupy":
                import cupy

                return cupy, partial(cupy.random.random, dtype=cupy.float64)
            elif backend == "torch":
                import torch

                return torch, partial(torch.rand, dtype=torch.float64)

        def _normal(backend, random, shape):
            theta = 2 * backend.pi * random(shape)
            r = backend.sqrt(-backend.log(random(shape)))
            z = r * (backend.cos(theta) + 1j * backend.sin(theta))
            return z

        backend, random = _backend()
        self.eta.data[:] = _normal(backend, random, self.eta.shape)

    @abstractmethod
    def sample(self, new_gauge: bool):
        pass

    @abstractmethod
    def action(self, new_gauge: bool) -> float:
        pass

    @abstractmethod
    def force(self, dt: float, new_gauge: bool):
        pass

    def _invertMultiShiftParam(self, mode: Literal["pseudo_fermion", "molecular_dynamics", "fermion_action"]):
        if mode == "pseudo_fermion":
            offset, residue, norm = (
                self.rational_param.offset_pseudo_fermion,
                self.rational_param.residue_pseudo_fermion,
                self.rational_param.norm_pseudo_fermion,
            )
        elif mode == "molecular_dynamics":
            offset, residue, norm = (
                self.rational_param.offset_molecular_dynamics,
                self.rational_param.residue_molecular_dynamics,
                None,
            )
        elif mode == "fermion_action":
            offset, residue, norm = (
                self.rational_param.offset_fermion_action,
                self.rational_param.residue_fermion_action,
                self.rational_param.norm_fermion_action,
            )
        assert len(offset) == len(residue)
        num_offset = len(offset)
        if num_offset > 1:
            tol = self.invert_param.tol
            self.invert_param.num_offset = num_offset
            self.invert_param.offset = offset + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
            self.invert_param.residue = residue + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
            self.invert_param.tol_offset = [
                max(tol * abs(residue[0] / residue[i]), 2e-16 * (self.latt_info.Gt * self.latt_info.Lt) ** 0.5)
                for i in range(num_offset)
            ] + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
        else:
            assert offset == [0.0] and residue == [1.0] and (norm is None or norm == 0.0)
        return residue, norm

    def _invertMultiShift(
        self,
        xx: Union[MultiLatticeFermion, MultiLatticeStaggeredFermion],
        x: Union[LatticeFermion, LatticeStaggeredFermion],
        b: Union[LatticeFermion, LatticeStaggeredFermion],
        residue: List[float],
        norm: float,
    ):
        num_offset = len(residue)
        if (
            self.invert_param.matpc_type == QudaMatPCType.QUDA_MATPC_EVEN_EVEN
            or self.invert_param.matpc_type == QudaMatPCType.QUDA_MATPC_EVEN_EVEN_ASYMMETRIC
        ):
            if num_offset > 1:
                invertMultiShiftQuda(xx.even_ptrs, b.even_ptr, self.invert_param)
                self.dirac.performance()
                if norm is not None:
                    x.even = norm * b.even
                    for i in range(num_offset):
                        x.even += residue[i] * xx[i].even
            else:
                if norm is None:
                    invertQuda(xx[0].even_ptr, b.even_ptr, self.invert_param)
                    self.dirac.performance()
                else:
                    self.invert_param.dagger = QudaDagType.QUDA_DAG_YES
                    MatQuda(x.even_ptr, b.even_ptr, self.invert_param)
                    self.invert_param.dagger = QudaDagType.QUDA_DAG_NO
        elif (
            self.invert_param.matpc_type == QudaMatPCType.QUDA_MATPC_ODD_ODD
            or self.invert_param.matpc_type == QudaMatPCType.QUDA_MATPC_ODD_ODD_ASYMMETRIC
        ):
            if num_offset > 1:
                invertMultiShiftQuda(xx.odd_ptrs, b.odd_ptr, self.invert_param)
                self.dirac.performance()
                if norm is not None:
                    x.odd = norm * b.odd
                    for i in range(num_offset):
                        x.odd += residue[i] * xx[i].odd
            else:
                if norm is None:
                    invertQuda(xx[0].odd_ptr, b.odd_ptr, self.invert_param)
                    self.dirac.performance()
                else:
                    self.invert_param.dagger = QudaDagType.QUDA_DAG_YES
                    MatQuda(x.odd_ptr, b.odd_ptr, self.invert_param)
                    self.invert_param.dagger = QudaDagType.QUDA_DAG_NO

    def invertMultiShift(
        self, mode: Literal["pseudo_fermion", "molecular_dynamics", "fermion_action"]
    ) -> Union[LatticeFermion, MultiLatticeFermion]:
        residue, norm = self._invertMultiShiftParam(mode)
        if mode == "pseudo_fermion":
            self._invertMultiShift(self.quark, self.phi, self.eta, residue, norm)
        else:
            self._invertMultiShift(self.quark, self.eta, self.phi, residue, norm)


class StaggeredFermionAction(FermionAction):
    dirac: StaggeredFermionDirac
    quark: MultiLatticeStaggeredFermion
    phi: LatticeStaggeredFermion
    eta: LatticeStaggeredFermion

    def __init__(self, latt_info: LatticeInfo, dirac: StaggeredFermionDirac) -> None:
        super().__init__(latt_info, dirac)
        self.is_staggered = True

    @abstractmethod
    def sample(self, new_gauge: bool):
        pass

    @abstractmethod
    def action(self, new_gauge: bool) -> float:
        pass

    @abstractmethod
    def force(self, dt: float, new_gauge: bool):
        pass

    def invertMultiShift(
        self, mode: Literal["pseudo_fermion", "molecular_dynamics", "fermion_action"]
    ) -> Union[LatticeStaggeredFermion, MultiLatticeStaggeredFermion]:
        residue, norm = self._invertMultiShiftParam(mode)
        if mode == "pseudo_fermion":
            self._invertMultiShift(self.quark, self.phi, self.eta, residue, norm)
        else:
            self._invertMultiShift(self.quark, self.eta, self.phi, residue, norm)



from typing import List, Literal, Union

import numpy

from pyquda_comm.field import (
    LatticeInfo,
    LatticeGauge,
    LatticeMom,
    LatticeFermion,
    LatticeStaggeredFermion,
    LatticeReal,
)
from ..pyquda import (
    QudaEigParam,
    QudaGaugeSmearParam,
    QudaGaugeObservableParam,
    MatQuda,
    computeGaugePathQuda,
    computeGaugeLoopTraceQuda,
    gaussGaugeQuda,
    gaussMomQuda,
    loadGaugeQuda,
    saveGaugeQuda,
    momResidentQuda,
    performWFlowQuda,
    freeUniqueGaugeQuda,
    staggeredPhaseQuda,
    performGaugeSmearQuda,
    performWuppertalnStep,
    gaugeObservablesQuda,
    projectSU3Quda,
    computeGaugeFixingOVRQuda,
    computeGaugeFixingFFTQuda,
)
from ..enum_quda import (
    QudaBoolean,
    QudaDslashType,
    QudaGaugeSmearType,
    QudaLinkType,
    QudaMassNormalization,
    QudaPrecision,
    QudaReconstructType,
    QudaSolveType,
)

from . import general
from .abstract import Dirac


class GaugeDirac(Dirac):
    def __init__(self, latt_info: LatticeInfo) -> None:
        super().__init__(LatticeInfo(latt_info.global_size))  # Keep periodic t boundary and isotropic
        self.newQudaGaugeParam()
        self.newQudaInvertParam()
        self.newQudaEigParam()
        self.newQudaGaugeSmearParam()
        self.newQudaGaugeObservableParam()
        self.setPrecision(
            cuda=QudaPrecision.QUDA_DOUBLE_PRECISION,
            sloppy=QudaPrecision.QUDA_DOUBLE_PRECISION,
            precondition=QudaPrecision.QUDA_DOUBLE_PRECISION,
            eigensolver=QudaPrecision.QUDA_DOUBLE_PRECISION,
        )
        # Use QUDA_RECONSTRUCT_NO to ensure slight deviations from SU(3) can be preserved
        self.setReconstruct(
            cuda=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            sloppy=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            precondition=QudaReconstructType.QUDA_RECONSTRUCT_NO,
            eigensolver=QudaReconstructType.QUDA_RECONSTRUCT_NO,
        )

    def newQudaGaugeParam(self):
        gauge_param = general.newQudaGaugeParam(self.latt_info, 1.0, 0.0)
        self.gauge_param = gauge_param

    def newQudaInvertParam(self):
        invert_param = general.newQudaInvertParam(0, 1 / 8, 0, 0, 0.0, 1.0, None)
        invert_param.solve_type = QudaSolveType.QUDA_DIRECT_SOLVE
        invert_param.mass_normalization = QudaMassNormalization.QUDA_KAPPA_NORMALIZATION
        self.invert_param = invert_param

    def newQudaEigParam(self):
        eig_param = QudaEigParam()
        eig_param.vec_infile = b""
        eig_param.vec_outfile = b""
        eig_param.invert_param = self.invert_param
        self.eig_param = eig_param

    def newQudaGaugeSmearParam(self):
        smear_param = QudaGaugeSmearParam()
        self.smear_param = smear_param

    def newQudaGaugeObservableParam(self):
        obs_param = QudaGaugeObservableParam()
        self.obs_param = obs_param

    def loadGauge(self, gauge: LatticeGauge):
        self.gauge_param.use_resident_gauge = 0
        loadGaugeQuda(gauge.data_ptrs, self.gauge_param)
        self.gauge_param.use_resident_gauge = 1

    def saveGauge(self, gauge: LatticeGauge):
        saveGaugeQuda(gauge.data_ptrs, self.gauge_param)

    def freeGauge(self):
        freeUniqueGaugeQuda(QudaLinkType.QUDA_WILSON_LINKS)

    def loadMom(self, mom: LatticeMom):
        momResidentQuda(mom.data_ptrs, self.gauge_param)

    def saveFreeMom(self, mom: LatticeMom):
        self.gauge_param.make_resident_mom = 0
        self.gauge_param.return_result_mom = 1
        momResidentQuda(mom.data_ptrs, self.gauge_param)
        self.gauge_param.make_resident_mom = 1
        self.gauge_param.return_result_mom = 0

    def saveSmearedGauge(self, gauge: LatticeGauge):
        self.gauge_param.type = QudaLinkType.QUDA_SMEARED_LINKS
        saveGaugeQuda(gauge.data_ptrs, self.gauge_param)
        self.gauge_param.type = QudaLinkType.QUDA_WILSON_LINKS

    def freeSmearedGauge(self):
        freeUniqueGaugeQuda(QudaLinkType.QUDA_SMEARED_LINKS)

    def covDev(self, x: LatticeFermion, covdev_mu: int):
        b = LatticeFermion(x.latt_info)
        self.invert_param.dslash_type = QudaDslashType.QUDA_COVDEV_DSLASH
        self.invert_param.mass = -3
        self.invert_param.kappa = 1 / 2
        self.invert_param.covdev_mu = covdev_mu
        MatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def laplace(self, x: LatticeStaggeredFermion, laplace3D: int):
        b = LatticeStaggeredFermion(x.latt_info)
        self.invert_param.dslash_type = QudaDslashType.QUDA_LAPLACE_DSLASH
        laplaceDim = 3 if laplace3D in [0, 1, 2, 3] else 4
        self.invert_param.mass = laplaceDim - 4
        self.invert_param.kappa = 1 / (2 * laplaceDim)
        self.invert_param.laplace3D = laplace3D
        MatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def wuppertalSmear(self, x: Union[LatticeFermion, LatticeStaggeredFermion], n_steps: int, alpha: float):
        if isinstance(x, LatticeStaggeredFermion):
            b = LatticeStaggeredFermion(x.latt_info)
            self.invert_param.dslash_type = QudaDslashType.QUDA_STAGGERED_DSLASH
        else:
            b = LatticeFermion(x.latt_info)
            self.invert_param.dslash_type = QudaDslashType.QUDA_WILSON_DSLASH
        performWuppertalnStep(b.data_ptr, x.data_ptr, self.invert_param, n_steps, alpha)
        return b

    def staggeredPhase(self, gauge: LatticeGauge, applied: bool):
        self.gauge_param.use_resident_gauge = 0
        self.gauge_param.make_resident_gauge = 0
        self.gauge_param.return_result_gauge = 1
        self.gauge_param.staggered_phase_applied = int(applied)
        staggeredPhaseQuda(gauge.data_ptrs, self.gauge_param)
        self.gauge_param.use_resident_gauge = 1
        self.gauge_param.make_resident_gauge = 1
        self.gauge_param.return_result_gauge = 0
        self.gauge_param.staggered_phase_applied = int(not applied)

    def projectSU3(self, gauge: LatticeGauge, tol: float):
        self.gauge_param.use_resident_gauge = 0
        self.gauge_param.make_resident_gauge = 0
        self.gauge_param.return_result_gauge = 1
        projectSU3Quda(gauge.data_ptrs, tol, self.gauge_param)
        self.gauge_param.use_resident_gauge = 1
        self.gauge_param.make_resident_gauge = 1
        self.gauge_param.return_result_gauge = 0

    @classmethod
    def _getPath(cls, path: List[int]):
        input_path_buf = numpy.zeros((len(path)), "<i4")
        for j, d in enumerate(path):
            if 0 <= d < 4:
                input_path_buf[j] = d
            elif 4 <= d < 8:
                input_path_buf[j] = 7 - (d - 4)
            else:
                raise ValueError(f"path should be list of int from 0 to 7, but get {path}")
        return input_path_buf, len(path)

    def path(
        self,
        gauge: LatticeGauge,
        paths: List[List[int]],
    ):
        gauge_path = LatticeGauge(gauge.latt_info)
        num_paths = 1
        input_path_buf_x, path_length = GaugeDirac._getPath(paths[0])
        input_path_buf = numpy.zeros((4, 1, path_length + 1), "<i4")
        input_path_buf[0, 0, 0] = 7
        input_path_buf[0, 0, 1:] = input_path_buf_x
        for d in range(1, 4):
            input_path_buf_, path_length_ = GaugeDirac._getPath(paths[d])
            assert path_length_ == path_length, "paths in all directions should have the same shape"
            input_path_buf[d, 0, 0] = 7 - d
            input_path_buf[d, 0, 1:] = input_path_buf_
        max_length = path_length
        path_length = numpy.array([path_length], "<i4")
        loop_coeff = numpy.ones((1), "<f8")
        self.gauge_param.overwrite_gauge = 1
        self.gauge_param.use_resident_gauge = 0
        self.gauge_param.make_resident_gauge = 0
        computeGaugePathQuda(
            gauge_path.data_ptrs,
            gauge.data_ptrs,
            input_path_buf,
            path_length + 1,
            loop_coeff,
            num_paths,
            max_length + 1,
            1.0,
            self.gauge_param,
        )
        self.gauge_param.overwrite_gauge = 0
        self.gauge_param.use_resident_gauge = 1
        self.gauge_param.make_resident_gauge = 1
        return gauge_path

    @classmethod
    def _getLoops(cls, loops: List[List[int]]):
        num_paths = len(loops)
        path_length = numpy.zeros((num_paths), "<i4")
        for i in range(num_paths):
            path_length[i] = len(loops[i])
        max_length = int(numpy.max(path_length))
        input_path_buf = numpy.full((num_paths, max_length), -1, "<i4")
        for i in range(num_paths):
            dx = [0, 0, 0, 0]
            for j, d in enumerate(loops[i]):
                if 0 <= d < 4:
                    dx[d] += 1
                    input_path_buf[i, j] = d
                elif 4 <= d < 8:
                    dx[d - 4] -= 1
                    input_path_buf[i, j] = 7 - (d - 4)
                else:
                    raise ValueError(f"path should be list of int from 0 to 7, but get {loops[i]}")
            if dx != [0, 0, 0, 0]:
                raise ValueError(f"path {loops[i]} is not a loop")
        return input_path_buf, path_length, num_paths, max_length

    def loop(
        self,
        gauge: LatticeGauge,
        loops: List[List[List[int]]],
        coeff: List[float],
    ):
        gauge_loop = LatticeGauge(gauge.latt_info)
        input_path_buf_x, path_length, num_paths, max_length = GaugeDirac._getLoops(loops[0])
        input_path_buf = numpy.zeros((4, num_paths, max_length + 1), "<i4")
        input_path_buf[0, :, 0] = 7
        input_path_buf[0, :, 1:] = input_path_buf_x
        for d in range(1, 4):
            input_path_buf_, path_length_, num_paths_, max_length_ = GaugeDirac._getLoops(loops[d])
            assert (path_length_ == path_length).all(), "paths in all directions should have the same shape"
            input_path_buf[d, :, 0] = 7 - d
            input_path_buf[d, :, 1:] = input_path_buf_
        loop_coeff = numpy.asarray(coeff, "<f8")
        self.gauge_param.overwrite_gauge = 1
        self.gauge_param.use_resident_gauge = 0
        self.gauge_param.make_resident_gauge = 0
        computeGaugePathQuda(
            gauge_loop.data_ptrs,
            gauge.data_ptrs,
            input_path_buf,
            path_length + 1,
            loop_coeff,
            num_paths,
            max_length + 1,
            1.0,
            self.gauge_param,
        )
        self.gauge_param.overwrite_gauge = 0
        self.gauge_param.use_resident_gauge = 1
        self.gauge_param.make_resident_gauge = 1
        return gauge_loop

    def loopTrace(self, loops: List[List[int]]):
        input_path_buf, path_length, num_paths, max_length = GaugeDirac._getLoops(loops)
        traces = numpy.zeros((num_paths), "<c16")
        loop_coeff = numpy.ones((num_paths), "<f8")
        computeGaugeLoopTraceQuda(
            traces,
            input_path_buf,
            path_length,
            loop_coeff,
            num_paths,
            max_length,
            1.0,
        )
        return traces

    def apeSmear(
        self,
        n_steps: int,
        alpha: float,
        dir_ignore: int,
        compute_plaquette: bool = False,
        compute_qcharge: bool = False,
    ):
        self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_APE
        self.smear_param.n_steps = n_steps
        self.smear_param.alpha = alpha
        self.smear_param.meas_interval = n_steps + 1
        self.smear_param.dir_ignore = dir_ignore
        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
        performGaugeSmearQuda(self.smear_param, self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE

    def stoutSmear(
        self,
        n_steps: int,
        rho: float,
        dir_ignore: int,
        compute_plaquette: bool = False,
        compute_qcharge: bool = False,
    ):
        self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_STOUT
        self.smear_param.n_steps = n_steps
        self.smear_param.rho = rho
        self.smear_param.meas_interval = n_steps + 1
        self.smear_param.dir_ignore = dir_ignore
        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
        performGaugeSmearQuda(self.smear_param, self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE

    def hypSmear(
        self,
        n_steps: int,
        alpha1: float,
        alpha2: float,
        alpha3: float,
        dir_ignore: int,
        compute_plaquette: bool = False,
        compute_qcharge: bool = False,
    ):
        self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_HYP
        self.smear_param.n_steps = n_steps
        self.smear_param.alpha1 = alpha1
        self.smear_param.alpha2 = alpha2
        self.smear_param.alpha3 = alpha3
        self.smear_param.meas_interval = n_steps + 1
        self.smear_param.dir_ignore = dir_ignore
        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
        performGaugeSmearQuda(self.smear_param, self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE

    def wilsonFlow(
        self,
        n_steps: int,
        epsilon: float,
        t0: float,
        restart: bool,
        compute_plaquette: bool = False,
        compute_qcharge: bool = True,
    ):
        self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_WILSON_FLOW
        self.smear_param.n_steps = n_steps
        self.smear_param.epsilon = epsilon
        self.smear_param.t0 = t0
        self.smear_param.restart = QudaBoolean(restart)
        self.smear_param.meas_interval = n_steps + 1
        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
        performWFlowQuda(self.smear_param, self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE

    def symanzikFlow(
        self,
        n_steps: int,
        epsilon: float,
        t0: float,
        restart: bool,
        compute_plaquette: bool = False,
        compute_qcharge: bool = True,
    ):
        self.smear_param.smear_type = QudaGaugeSmearType.QUDA_GAUGE_SMEAR_SYMANZIK_FLOW
        self.smear_param.n_steps = n_steps
        self.smear_param.epsilon = epsilon
        self.smear_param.t0 = t0
        self.smear_param.restart = QudaBoolean(restart)
        self.smear_param.meas_interval = n_steps + 1
        self.obs_param.compute_plaquette = QudaBoolean(compute_plaquette)
        self.obs_param.compute_qcharge = QudaBoolean(compute_qcharge)
        performWFlowQuda(self.smear_param, self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE

    def plaquette(self):
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.plaquette

    def polyakovLoop(self):
        self.obs_param.compute_polyakov_loop = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_polyakov_loop = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.ploop

    def energy(self):
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.energy

    def qcharge(self):
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.qcharge

    def qchargeDensity(self):
        qcharge_density = LatticeReal(self.latt_info)
        self.obs_param.qcharge_density = qcharge_density.data_ptr
        self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
        return qcharge_density

    def gaussGauge(self, seed: int, sigma: float):
        gaussGaugeQuda(seed, sigma)

    def gaussMom(self, seed: int, sigma: float):
        gaussMomQuda(seed, sigma)

    def fixingOVR(
        self,
        gauge: LatticeGauge,
        gauge_dir: Literal[3, 4],
        Nsteps: int,
        verbose_interval: int,
        relax_boost: float,
        tolerance: float,
        reunit_interval: int,
        stopWtheta: int,
    ):
        computeGaugeFixingOVRQuda(
            gauge.data_ptrs,
            gauge_dir,
            Nsteps,
            verbose_interval,
            relax_boost,
            tolerance,
            reunit_interval,
            stopWtheta,
            self.gauge_param,
        )

    def fixingFFT(
        self,
        gauge: LatticeGauge,
        gauge_dir: Literal[3, 4],
        Nsteps: int,
        verbose_interval: int,
        alpha: float,
        autotune: int,
        tolerance: float,
        stopWtheta: int,
    ):
        computeGaugeFixingFFTQuda(
            gauge.data_ptrs,
            gauge_dir,
            Nsteps,
            verbose_interval,
            alpha,
            autotune,
            tolerance,
            stopWtheta,
            self.gauge_param,
        )



from time import perf_counter
from typing import List, Sequence, Tuple, Union

import numpy
from mpi4py import MPI
import h5py

from . import getLogger, getMPIComm, getGridSize, getGridCoord


class _LatticeInfo:
    def __init__(self, latt_size: List[int]) -> None:
        grid_size = getGridSize()
        self._checkLattice(latt_size, grid_size)
        self._setLattice(latt_size, grid_size)

    def _checkLattice(self, latt_size: List[int], grid_size: List[int]):
        assert len(latt_size) == len(grid_size), "lattice size and grid size must have the same dimension"
        for GL, G in zip(latt_size, grid_size):
            if not (GL % G == 0):
                getLogger().critical(f"lattice size {latt_size} must be divisible by gird size {grid_size}", ValueError)

    def _setLattice(self, latt_size: List[int], grid_size: List[int]):
        grid_coord = getGridCoord()
        sublatt_size = [GL // G for GL, G in zip(latt_size, grid_size)]
        sublatt_slice = []
        for g, L in zip(grid_coord[::-1], sublatt_size[::-1]):
            sublatt_slice.append(slice(g * L, (g + 1) * L))

        self.global_size = latt_size
        self.global_volume = int(numpy.prod(latt_size))
        self.size = sublatt_size
        self.volume = int(numpy.prod(sublatt_size))
        self.slice = tuple(sublatt_slice)


# CRC32LUT = numpy.empty((4, 256), dtype="<u4")
# # fmt: off
# CRC32LUT[0] = numpy.array(
#     [
#         0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA, 0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3,
#         0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988, 0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91,
#         0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE, 0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7,
#         0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC, 0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5,
#         0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172, 0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B,
#         0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940, 0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59,
#         0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116, 0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F,
#         0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924, 0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D,
#         0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A, 0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433,
#         0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818, 0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01,
#         0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E, 0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457,
#         0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C, 0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65,
#         0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2, 0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB,
#         0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0, 0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9,
#         0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086, 0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F,
#         0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4, 0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD,
#         0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A, 0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683,
#         0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8, 0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1,
#         0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE, 0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7,
#         0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC, 0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5,
#         0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252, 0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B,
#         0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60, 0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79,
#         0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236, 0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F,
#         0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04, 0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D,
#         0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A, 0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713,
#         0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38, 0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21,
#         0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E, 0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777,
#         0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C, 0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45,
#         0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2, 0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB,
#         0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0, 0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9,
#         0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6, 0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF,
#         0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94, 0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D
#     ],
#     "<u4"
# )
# # fmt: on
# for i in range(256):
#     CRC32LUT[1, i] = (CRC32LUT[0, i] >> 8) ^ CRC32LUT[0, CRC32LUT[0, i] & 0xFF]
#     CRC32LUT[2, i] = (CRC32LUT[1, i] >> 8) ^ CRC32LUT[0, CRC32LUT[1, i] & 0xFF]
#     CRC32LUT[3, i] = (CRC32LUT[2, i] >> 8) ^ CRC32LUT[0, CRC32LUT[2, i] & 0xFF]
#     # CRC32LUT[4, i] = (CRC32LUT[3, i] >> 8) ^ CRC32LUT[0, CRC32LUT[3, i] & 0xFF]
#     # CRC32LUT[5, i] = (CRC32LUT[4, i] >> 8) ^ CRC32LUT[0, CRC32LUT[4, i] & 0xFF]
#     # CRC32LUT[6, i] = (CRC32LUT[5, i] >> 8) ^ CRC32LUT[0, CRC32LUT[5, i] & 0xFF]
#     # CRC32LUT[7, i] = (CRC32LUT[6, i] >> 8) ^ CRC32LUT[0, CRC32LUT[6, i] & 0xFF]


def checksum(latt_info, data: numpy.ndarray) -> Tuple[int, int]:
    # import zlib

    # work = numpy.empty((latt_info.volume), "<u4")
    # for i in range(latt_info.volume):
    #     work[i] = zlib.crc32(data[i])
    # work = numpy.full_like(data[:, 0], 0xFFFFFFFF)
    # for i in range(data.shape[1]):
    #     work ^= data[:, i]
    #     work_view = work.view("|u1").reshape(-1, 4)
    #     work = (
    #         CRC32LUT[0].take(work_view[:, 0], mode="wrap")
    #         ^ CRC32LUT[1].take(work_view[:, 1], mode="wrap")
    #         ^ CRC32LUT[2].take(work_view[:, 2], mode="wrap")
    #         ^ CRC32LUT[3].take(work_view[:, 3], mode="wrap")
    #     )
    # work ^= 0xFFFFFFFF
    work = numpy.bitwise_xor.reduce(data.view("<u4"), 1)
    rank = (
        numpy.arange(latt_info.global_volume, dtype="<u8")
        .reshape(*latt_info.global_size[::-1])[latt_info.slice]
        .reshape(-1)
    )
    rank29 = (rank % 29).astype("<u4")
    rank31 = (rank % 31).astype("<u4")
    sum29 = getMPIComm().allreduce(numpy.bitwise_xor.reduce(work << rank29 | work >> (32 - rank29)), MPI.BXOR)
    sum31 = getMPIComm().allreduce(numpy.bitwise_xor.reduce(work << rank31 | work >> (32 - rank31)), MPI.BXOR)
    return sum29, sum31


def _spin_color_dtype(name: str, shape: Sequence[int], use_fp32: bool = True) -> Tuple[int, int]:
    float_nbytes = 4 if use_fp32 else 8
    Ns, Nc, dtype = 4, 3, f"<c{2 * float_nbytes}"
    if name.endswith("Int"):
        () = shape
        dtype = "<i4"
    elif name.endswith("Real"):
        () = shape
        dtype = f"<f{float_nbytes}"
    elif name.endswith("Complex"):
        () = shape
    elif name.endswith("SpinColorVector"):
        Ns, Nc = shape
    elif name.endswith("SpinColorMatrix"):
        Ns, Ns_, Nc, Nc_ = shape
        assert Ns == Ns_ and Nc == Nc_
    elif name.endswith("ColorVector"):
        (Nc,) = shape
    elif name.endswith("ColorMatrix"):
        Nc, Nc_ = shape
        assert Nc == Nc_
    else:
        raise ValueError(f"Invalid field type: {name}")
    return Ns, Nc, dtype


def _field_info(group: str, label: Union[int, str, Sequence[str], Sequence[str]], field: numpy.ndarray, use_fp32: bool):
    grid_size = getGridSize()
    Nd = len(grid_size)
    if isinstance(label, (int, str)):
        keys = str(label)
        sublatt_size = field.shape[0:Nd][::-1]
        field_shape = field.shape[Nd:]
    elif isinstance(label, (list, tuple, range)):
        assert len(label) == field.shape[0]
        keys = [str(key) for key in label]
        sublatt_size = field.shape[1 : 1 + Nd][::-1]
        field_shape = field.shape[1 + Nd :]
    else:
        raise TypeError(f"Invalid label {label} for field type {group}")
    latt_size = [G * L for G, L in zip(grid_size, sublatt_size)]
    Ns, Nc, field_dtype = _spin_color_dtype(group, field_shape, use_fp32)
    return keys, latt_size, Ns, Nc, field_shape, field_dtype


class File(h5py.File):
    def __init__(self, name, mode="r", **kwds):
        """Create a new file object with the mpio driver.

        See the h5py user guide for a detailed explanation of the options.

        name
            Name of the file on disk, or file-like object.
        mode
            r        Readonly, file must exist (default)
            r+       Read/write, file must exist
            w        Create file, truncate if exists
            w- or x  Create file, fail if exists
            a        Read/write if exists, create otherwise
        """
        super().__init__(name, mode, driver="mpio", comm=getMPIComm(), **kwds)

    @classmethod
    def _load(cls, latt_info: _LatticeInfo, dataset: h5py.Dataset, check: bool = True):
        data: numpy.ndarray = dataset[latt_info.slice]
        if check:
            sum29, sum31 = checksum(latt_info, data.reshape(latt_info.volume, -1))
            assert dataset.attrs["sum29"] == f"{sum29:08x}", f"{dataset.attrs['sum29']} != {sum29:08x}"
            assert dataset.attrs["sum31"] == f"{sum31:08x}", f"{dataset.attrs['sum31']} != {sum31:08x}"
        return data

    def load(self, group: str, label: Union[int, str, Sequence[int], Sequence[str]], *, check: bool = True):
        s = perf_counter()
        gbytes = 0
        g = self[group]
        # assert g.attrs["Lattice"] == " ".join([str(L) for L in latt_size])
        # assert g.attrs["Spin"] == str(Ns)
        # assert g.attrs["Color"] == str(Nc)
        latt_size = [int(GL) for GL in g.attrs["Lattice"].split()]
        Ns = int(g.attrs["Spin"]) if "Spin" in group else None
        Nc = int(g.attrs["Color"]) if "Color" in group else None
        if isinstance(label, (int, str)):
            keys = str(label)
        elif isinstance(label, (list, tuple, range)):
            assert len(label) <= len(g)
            keys = [str(key) for key in label]

        for key in g.keys():
            field_dtype = g[key].dtype.str.replace("<c8", "<c16").replace("<f4", "<f8")
            break
        latt_info = _LatticeInfo(latt_size)
        if isinstance(keys, str):
            key = keys
            value = self._load(latt_info, g[key], check).astype(field_dtype)
            gbytes += g[key].nbytes / 1024**3
        else:
            value = []
            for key in keys:
                value.append(self._load(latt_info, g[key], check).astype(field_dtype))
                gbytes += g[key].nbytes / 1024**3
        secs = perf_counter() - s
        getLogger().info(f"Loaded {group} from {self.filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")
        return latt_size, Ns, Nc, value

    @classmethod
    def _save(cls, latt_info: _LatticeInfo, dataset: h5py.Dataset, data: numpy.ndarray, check: bool = True):
        dataset[latt_info.slice] = data
        if check:
            sum29, sum31 = checksum(latt_info, data.reshape(latt_info.volume, -1))
            dataset.attrs["sum29"] = f"{sum29:08x}"
            dataset.attrs["sum31"] = f"{sum31:08x}"

    def save(
        self,
        group: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        field: numpy.ndarray,
        *,
        annotation: str = "",
        check: bool = True,
        use_fp32: bool = False,
    ):
        s = perf_counter()
        gbytes = 0
        keys, latt_size, Ns, Nc, field_shape, field_dtype = _field_info(group, label, field, use_fp32)
        g = self.create_group(group)
        g.attrs["Annotation"] = annotation
        g.attrs["Lattice"] = " ".join([str(GL) for GL in latt_size])
        if "Spin" in group:
            g.attrs["Spin"] = str(Ns)
        if "Color" in group:
            g.attrs["Color"] = str(Nc)

        latt_info = _LatticeInfo(latt_size)
        if isinstance(keys, str):
            key = keys
            g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
            self._save(latt_info, g[key], field.astype(field_dtype), check)
            gbytes += g[key].nbytes / 1024**3
        else:
            for index, key in enumerate(keys):
                g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
                self._save(latt_info, g[key], field[index].astype(field_dtype), check)
                gbytes += g[key].nbytes / 1024**3
        secs = perf_counter() - s
        getLogger().info(f"Saved {group} to {self.filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    def append(
        self,
        group: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        field: numpy.ndarray,
        *,
        annotation: str = "",
        check: bool = True,
        use_fp32: bool = False,
    ):
        s = perf_counter()
        gbytes = 0
        keys, latt_size, Ns, Nc, field_shape, field_dtype = _field_info(group, label, field, use_fp32)
        g = self.create_group(group)
        g.attrs["Annotation"] = annotation
        g.attrs["Lattice"] = " ".join([str(GL) for GL in latt_size])
        if "Spin" in group:
            g.attrs["Spin"] = str(Ns)
        if "Color" in group:
            g.attrs["Color"] = str(Nc)

        latt_info = _LatticeInfo(latt_size)
        if isinstance(keys, str):
            key = keys
            g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
            self._save(latt_info, g[key], field.astype(field_dtype), check)
            gbytes += g[key].nbytes / 1024**3
        else:
            for index, key in enumerate(keys):
                g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
                self._save(latt_info, g[key], field[index].astype(field_dtype), check)
                gbytes += g[key].nbytes / 1024**3
        secs = perf_counter() - s
        getLogger().info(f"Append {group} to {self.filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    def update(
        self,
        group: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        field: numpy.ndarray,
        *,
        annotation: str = "",
        check: bool = True,
    ):
        s = perf_counter()
        gbytes = 0
        keys, latt_size, Ns, Nc, field_shape, field_dtype = _field_info(group, label, field, False)
        g = self[group]
        if annotation != "":
            g.attrs["Annotation"] = annotation
        assert g.attrs["Lattice"] == " ".join([str(L) for L in latt_size])
        if "Spin" in group:
            assert g.attrs["Spin"] == str(Ns)
        if "Color" in group:
            assert g.attrs["Color"] == str(Nc)

        for key in g.keys():
            field_dtype = g[key].dtype.str
            break
        latt_info = _LatticeInfo(latt_size)
        if isinstance(keys, str):
            key = keys
            if key not in g:
                g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
            self._save(latt_info, g[key], field.astype(field_dtype), check)
            gbytes += g[key].nbytes / 1024**3
        else:
            for index, key in enumerate(keys):
                if key not in g:
                    g.create_dataset(key, (*latt_size[::-1], *field_shape), field_dtype)
                self._save(latt_info, g[key], field[index].astype(field_dtype), check)
                gbytes += g[key].nbytes / 1024**3
        secs = perf_counter() - s
        getLogger().info(f"Updated {group} to {self.filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")



from enum import IntEnum

# quda_constants.py

QUDA_INVALID_ENUM = -0x7FFFFFFF - 1


class qudaError_t(IntEnum):
    QUDA_SUCCESS = 0
    QUDA_ERROR = 1
    QUDA_ERROR_UNINITIALIZED = 2


class QudaDslashType(IntEnum):
    """
    Note: make sure QudaDslashType has corresponding entries in
    tests/utils/misc.cpp
    """

    pass


class QudaEigSpectrumType(IntEnum):
    """
    S=smallest L=largest\n
    R=real M=modulus I=imaniary
    """

    pass


class QudaCABasis(IntEnum):
    """Which basis to use for CA algorithms"""

    pass


class QudaMatPCType(IntEnum):
    """
    Whether the preconditioned matrix is (1-k^2 Deo Doe) or (1-k^2 Doe Deo)

    For the clover-improved Wilson Dirac operator, QUDA_MATPC_EVEN_EVEN
    defaults to the "symmetric" form, (1 - k^2 A_ee^-1 D_eo A_oo^-1 D_oe)
    and likewise for QUDA_MATPC_ODD_ODD.

    For the "asymmetric" form, (A_ee - k^2 D_eo A_oo^-1 D_oe), select
    QUDA_MATPC_EVEN_EVEN_ASYMMETRIC.
    """

    pass


class QudaParity(IntEnum):
    """Type used for "parity" argument to dslashQuda()"""

    pass


class QudaFieldLocation(IntEnum):
    """Where the field is stored"""

    pass


class QudaSiteSubset(IntEnum):
    """Which sites are included"""

    pass


class QudaSiteOrder(IntEnum):
    """Site ordering (always t-z-y-x, with rightmost varying fastest)"""

    pass


class QudaFieldOrder(IntEnum):
    """Degree of freedom ordering"""

    pass


class QudaGammaBasis(IntEnum):
    """
    gamj=((top 2 rows)(bottom 2 rows))  s1,s2,s3 are Pauli spin matrices, 1 is 2x2 identity

    Dirac-Pauli -> DeGrand-Rossi   T = i/sqrt(2)*((s2,-s2)(s2,s2))     field_DR = T * field_DP\n
    UKQCD -> DeGrand-Rossi         T = i/sqrt(2)*((-s2,-s2)(-s2,s2))   field_DR = T * field_UK\n
    Chiral -> DeGrand-Rossi        T = i*((0,-s2)(s2,0))               field_DR = T * field_chiral
    """

    pass


class QudaProjectionType(IntEnum):
    """used to select projection method for deflated solvers"""

    pass


class QudaPCType(IntEnum):
    """used to select checkerboard preconditioning method"""

    pass


class QudaExtLibType(IntEnum):
    """Allows to choose an appropriate external library"""

    pass
